{{
ParallelDevice.spin - Parallel Propeller Device communications driver
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below. 
}}

con
  CMD_RDBUF     = 0             ' host cmd -1
  CMD_WRBUF     = 1
  CMD_RUNRPC    = 2
  CMD_REGRPC    = 3
  CMD_NOP       = 4
  
obj
  def: "ParallelDefs"

var
  long cog
{
  long  command
  long  table
  long  destination
  long  address
  long  length
  long  bufferptr
  long  propid
}
    
pub start(parmp)
  stop
  long[parmp+4] := @cmdtable
  long[parmp] := CMD_NOP
  cog := cognew(@entry, parmp) + 1
  repeat while long[parmp]
  return cog

pub stop
  if cog
    cogstop(cog~ -1)

dat
'----------------------------------------------------------------------
entry          org      0

  jmp          #init

' pointers in this order
'
cmdptr         long 0   ' COG interface command address
tabptr         long 0   ' COG interface command table pointer address
dstptr         long 0   ' COG interface destination address
adrptr         long 0   ' COG interface address address
lenptr         long 0   ' COG interface buffer length address
bufptr         long 0   ' COG interface buffer data address
pidptr         long 0   ' COG interface Propeller ID address

cmd            long 0   ' one ofcommand table index commands
tab            long 0   ' command table commands
dst            long 0   ' the Secondary Propeller ID and COG target
adr            long 0   ' command address
len            long 0   ' length of buffer to send / receive
buf            long 0   ' buffer start
pid            long 0   ' Propeller Chip ID

tp             long 0   ' temporary pointer
t0             long 0   ' temporary variable
t1             long 0   ' temporary variable
t2             long 0   ' temporary variable

ack            long 1 << def#ACKPIN     ' always input  for host/device
atn            long 1 << def#ATNPIN     ' always output for host/device
dbus           long $ff                 ' output at command phase. in/out on data.
dbusack        long $ff | (1<<def#ACKPIN)

'----------------------------------------------------------------------
init
  mov          tp,      par
  mov          cmdptr,  tp              ' cmd changes always
  add          tp,      #4
  mov          tabptr,  tp              ' only need to get this once
  add          tp,      #4
  mov          dstptr,  tp              ' dst changes always
  add          tp,      #4
  mov          adrptr,  tp              ' adr changes always
  add          tp,      #4
  mov          lenptr,  tp              ' len changes always
  add          tp,      #4
  mov          bufptr,  tp              ' buf changes always
  add          tp,      #4
  rdlong       pid,     tp              ' our propeller chip id
  and          pid,     #7

  'mov          dira,    #0              ' always input
    
'----------------------------------------------------------------------
nextcmd        ' do next cmd
  andn         dira,    dbusack         ' disassert dbus and ack
  test         atn,     ina wz          ' wait for not attention  
  if_nz jmp    #nextcmd                 ' nothing to do 
  andn         outa,    ack             ' clear ack to let host know we got it
  mov          dst,     ina             ' get BYTE0
  mov          cmd,     dst             ' copy bits ... propid in bits[0:2]
  and          dst,     #$7             ' check only chip id
  cmp          dst,     pid wz          ' is message for me
  if_nz jmp    #nextcmd                 ' not for me 
  movs         dst,     ina             ' its for me ... get BYTE0
  shr          cmd,     #3              ' cogid in bits[0:2]
  mov          dst,     cmd             ' copy cogid
  and          dst,     #7              ' mask it ... cog id only please
  shr          cmd,     #3              ' command in bits[0:1]
  and          cmd,     #3              ' make sure only bits[0:1] are set
  shl          cmd,     #1              ' word wide table
  rdlong       tp,      tabptr          ' get command table
  add          tp,      cmd             ' add offset for command
  rdword       cmd,     tp              ' get command address
  mov          tp,      #0              ' zero values
  mov          t0,      #0              ' zero values
  mov          len,     #0              ' zero values
  or           dira,    ack             ' assert ack
  waitpeq      atn,     atn             ' wait for host to release atn  
  jmp          cmd                      ' jump to command, returns to nextcmd
  
'----------------------------------------------------------------------
_idle          ' do nothing command used only at startup 
  jmp          #nextcmd

'----------------------------------------------------------------------
' Host asserts ATT for any command to get device attention.
' First ATT byte contains cmd request, device id, and cog id
' Next contains addr low, then next contains addr high.
' Last header byte contains length of data.
' Rest of bytes contain data.
'
' BYTE0  CMDREQ[2],COGID[3],DEVID[3]
' BYTE1  ADDRLOW[8]
' BYTE2  ADDRHIGH[8]
' BYTE3  LENGTH[8]
' BYTE4  DATA0..N ... for RPC commands, this data is a packet
'
' For header, ACK low means device is ready.
' For reads, ACK high signals byte ready or write complete.
' For writes, RWN signals byte ready.   
'

'----------------------------------------------------------------------
_rdbufd        ' read from buffer / write to host
  movs         tp,      ina     ' BYTE1 - low address
  movs         t0,      ina     ' BYTE2 - high address
  movs         len,     ina     ' BYTE3 - data length
  and          tp,      #$ff    ' ............. mov rst release to bit 9 please
  and          t0,      #$ff    ' mask ... mov reset release out of bit 8
  and          len,     #$ff    ' ............. mov rst release to bit 9 please
  shl          t0,      #8      ' compile address
  or           tp,      t0      ' target address
  or           dira,    dbus    ' read from hub, write to host
:rlp rdbyte    outa,    tp      ' read data
  nop
  or           outa,    ack     ' say data good
  add          tp,      #1      ' increment address
  nop
  andn         outa,    ack     ' clear ack
  djnz         len,     #:rlp   ' not necessary ?
  jmp          #nextcmd

'----------------------------------------------------------------------
_wrbufd        ' read from host / write to buffer
  movs         tp,      ina     ' BYTE1 ... addr low
  movs         t0,      ina     ' BYTE2 ... addr high
  movs         len,     ina     ' BYTE3
  and          len,     #$ff    ' ............. mov rst release to bit 9 please
  and          tp,      #$ff    ' mask ... mov reset release out of bit 8
  and          t0,      #$ff    ' mask ... mov reset release out of bit 8
  shl          t0,      #8      ' compile address
  or           tp,      t0      ' target address
  andn         outa,    ack     ' clear ack
  andn         dira,    dbus    ' read
:wlp or        outa,    ack     ' say data good
  nop                           ' wait for host  
  mov          t0,      ina     ' get data
  wrbyte       t0,      tp      ' write data
  add          tp,      #1      ' increment address
  andn         outa,    ack     ' clear ack
  djnz         len,     #:wlp   ' not necessary ?
  jmp          #nextcmd

'----------------------------------------------------------------------
_runrpcd       ' execute a previously arranged remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
_regrpcd       ' register a remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
dat     ' debug command table
  long         ' align to long
  cmdtable     word _rdbufd     ' 0 CMD_RDBUF    - read from buffer to target
               word _wrbufd     ' 1 CMD_WRBUF    - write to buffer from target
               word _runrpcd    ' 2 CMD_RUN_RPC  - execute a remote function
               word _regrpcd    ' 3 CMD_REGISTER - register a remote function
               word _idle       ' 4 CMD_NOP      - used only at startup
              
{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}