{{
ParallelHost.spin - Parallel Propeller Host communications driver
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below. 
}}

obj
  def: "ParallelDefs"
  
var
  long cog
{
  long  command
  long  table
  long  destination
  long  address
  long  length
  long  bufferptr
}

'{    
pub start(parmp)
  stop
  long[parmp+4] := @cmdtable
  long[parmp] := def#CMD_NOP
  cog := cognew(@entry, parmp) + 1
  repeat while long[parmp]
  return cog
'}

{
obj
  v : "SpudVMM"

pub start(parmp)
  v.copyPasmStub(@entry)
  stop
  long[parmp] := def#CMD_NOP
  cog := v.startPASM(@entry, parmp) + 1
  repeat while long[parmp]
  return cog
'}    

pub stop
  if cog
    cogstop(cog~ -1)

pub getPasm
  return @entry
  
pub rdbuffer(parmp)
  long[parmp] := def#CMD_RDBUF
  repeat while long[parmp]
  
pub wrbuffer(parmp)
  long[parmp] := def#CMD_WRBUF
  repeat while long[parmp]
  
dat
'----------------------------------------------------------------------
entry          org      0
  long         0 [7] ' nops will be replaced by PASM debugger stub
  jmp          #init

' pointers in this order
'
cmdptr         long 0   ' COG interface command address
tabptr         long 0   ' COG interface command table pointer address
dstptr         long 0   ' COG interface destination address
adrptr         long 0   ' COG interface address address
lenptr         long 0   ' COG interface buffer length address
bufptr         long 0   ' COG interface buffer data address

cmd            long 0   ' one ofcommand table index commands
tab            long 0   ' command table commands
dst            long 0   ' the Secondary Propeller ID and COG target
adr            long 0   ' command address
len            long 0   ' length of buffer to send / receive
buf            long 0   ' buffer start

tp             long 0   ' temporary pointer
t0             long 0   ' temporary variable
t1             long 0   ' temporary variable
t2             long 0   ' temporary variable

ack            long 1 << def#ACKPIN     ' always input  for host/device
atn            long 1 << def#ATNPIN     ' always output for host/device
dbus           long $ff                 ' output at command phase. in/out on data.
zero           long 0

'----------------------------------------------------------------------
init           'jmp      #init
  mov          tp,      par
  mov          cmdptr,  tp              ' cmd changes always
  add          tp,      #4
  mov          tabptr,  tp              ' only need to get this once
  add          tp,      #4
  mov          dstptr,  tp              ' dst changes always
  add          tp,      #4
  mov          adrptr,  tp              ' adr changes always
  add          tp,      #4
  mov          lenptr,  tp              ' len changes always
  add          tp,      #4
  mov          bufptr,  tp              ' buf changes always

'----------------------------------------------------------------------
nextcmd        ' do next cmd
  or           dira,    dbus            ' start command with output enabled
  wrlong       zero,    cmdptr          ' let user know command is done
:loop
  rdlong       cmd,     cmdptr          ' get command
  tjz          cmd,     #:loop          ' wait for valid command
  rdlong       dst,     dstptr          ' get destination
  rdlong       adr,     adrptr          ' get target address
  rdlong       tp,      tabptr          ' get command table address
  shl          cmd,     #1              ' commands are word aligned
  add          tp,      cmd             ' add offset for command
  rdlong       len,     lenptr          ' get buffer length
  shr          cmd,     #1              ' command bits are byte aligned
  sub          cmd,     #1              ' commands to device are CMD-1 ... 2 bits
  rdlong       buf,     bufptr          ' get buffer data pointer
  shl          cmd,     #6              ' mov cmd to BYTE0 CMDREQ[2]
  or           cmd,     dst             ' BYTE0
  mov          outa,    cmd             ' BYTE0 set destination, attention, and cmd
  andn         outa,    atn     ' got response, disassert attention
  rdword       t0,      tp              ' get command address
  jmp          t0                       ' jump to command, returns to nextcmd

'----------------------------------------------------------------------
_idle          ' do nothing command used only at startup 
  jmp          #nextcmd

'----------------------------------------------------------------------
' Host asserts ATT for any command to get device attention.
' First ATT byte contains cmd request, device id, and cog id
' Next contains addr low, then next contains addr high.
' Last header byte contains length of data.
' Rest of bytes contain data.
'
' BYTE0  CMDREQ[2],COGID[3],DEVID[3]
' BYTE1  ADDRLOW[8]
' BYTE2  ADDRHIGH[8]
' BYTE3  LENGTH[8]
' BYTE4  DATA0..N ... for RPC commands, this data is a packet
'
' For header, ACK low means device is ready.
' For reads, ACK high signals byte ready or write complete.
' For writes, RWN signals byte ready.   
'

'----------------------------------------------------------------------
_rdbuf         ' read buffer from target
  or           dira,    atn     ' assert attention
  waitpne      ack,     ack     ' wait for ack low
  or           outa,    atn     ' got response, disassert attention
  andn         dira,    atn     ' release attention
  mov          outa,    adr     ' BYTE1
  shr          outa,    #8      ' BYTE2
  mov          outa,    len     ' BYTE3
  andn         dira,    dbus
:lp waitpeq    ack,     ack     ' wait for ack high ... device is writing
  mov          t0,      ina     ' BYTE4+n ... get byte
  wrbyte       t0,      buf     ' save data to hub
  add          buf,     #1      ' increment address
  djnz         len,     #:lp    ' not necessary ?
  jmp          #nextcmd

'----------------------------------------------------------------------
_wrbuf         ' write buffer to target
  or           dira,    atn     ' assert attention
  waitpne      ack,     ack     ' wait for ack low
  or           outa,    atn     ' got response, disassert attention
  andn         dira,    atn     ' release attention
  mov          outa,    adr     ' BYTE1
  shr          outa,    #8      ' BYTE2
  mov          outa,    len     ' BYTE3
  waitpne      ack,     ack     ' wait for ack low
:lp rdbyte     outa,    buf     ' BYTE4+n ... set byte
  add          buf,     #1      ' increment address
  waitpeq      ack,     ack     ' wait for ack high ... device is reading
  djnz         len,     #:lp    ' not necessary ?
  jmp          #nextcmd

'----------------------------------------------------------------------
_runrpc        ' execute a previously arranged remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
_regrpc        ' register a remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
dat     ' debug command table
  long         ' align to long
  cmdtable     word _idle       ' 0              - IDLE command
               word _rdbuf      ' 1 def#CMD_RDBUF    - read buffer from target
               word _wrbuf      ' 2 def#CMD_WRBUF    - write buffer to target
               word _runrpc     ' 3 def#CMD_RUN_RPC  - execute a remote function
               word _regrpc     ' 4 def#CMD_REGISTER - register a remote function
               word _idle       ' 5 def#CMD_NOP      - used only at startup
              
{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}